import type { ESLint } from 'eslint';
import * as fs from 'node:fs';
import { resolve, basename, extname } from 'node:path';
import { glob } from 'fast-glob';
import { parseModule } from 'esprima';
import query from 'esquery';
import type { Node, Property } from 'estree';

// https://github.com/gulpjs/interpret
const transpilers = [
  'esbin/register',
  'esbuild-register',
  'ts-node/register/transpile-only',
  '@swc/register',
  'sucrase/register',
  '@babel/register',
  'coffeescript/register',
];

function tryRequire() {
  for (const candidate of transpilers) {
    try {
      require(candidate);
      return;
    } catch {}
  }
}

const unwrapDefault = <T = any>(module: any): T => module.default ?? module;

const plugin: ESLint.Plugin = {
  rules: {},
};

function hydrateESTreeNode(n: Node): any {
  switch (n.type) {
    case 'Literal':
      return n.value;
    case 'ArrayExpression':
      return n.elements.filter(Boolean).map(hydrateESTreeNode);
    default:
      throw new Error(`Unsupported node type: ${n.type}`);
  }
}

function parseConfigFile(js: string) {
  const [node] = query(
    parseModule(js),
    'CallExpression[callee.name="extendConfig"] > ObjectExpression > Property[key.name="customRuleFiles"]',
  );
  return hydrateESTreeNode((node as Property).value);
}

function main() {
  const rootDir = process.cwd();

  const eslintConfigFile = ['.eslintrc.js', '.eslintrc.cjs']
    .map(file => resolve(rootDir, file))
    .find(file => fs.existsSync(file));

  if (!eslintConfigFile) return;

  const eslintConfig = fs.readFileSync(eslintConfigFile, 'utf8');
  const customRuleFiles = parseConfigFile(eslintConfig);
  if (!customRuleFiles?.length) return;

  tryRequire();
  for (const file of glob.sync(customRuleFiles)) {
    const module = unwrapDefault(require(file));
    const name = module.name ?? basename(file, extname(file));
    plugin.rules![name] = module;
  }
}

main();

export = plugin;
